"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.evmLogSubscriptionHandler = void 0;
const SocketManager_1 = require("./SocketManager");
const validateBaseRequest = (request) => {
    if (typeof request !== 'object' || request === null) {
        return [false, 'Request must be an object.'];
    }
    const req = request;
    if (typeof req.method !== 'string') {
        return [false, 'method must be a string.'];
    }
    return [true, ''];
};
const validateSubscribeRequest = (request) => {
    if (typeof request !== 'object' || request === null) {
        return [false, 'Request must be an object.'];
    }
    const req = request;
    if (typeof req.params !== 'object' || req.params === null) {
        return [false, 'params must be an object.'];
    }
    const params = req.params;
    if (typeof params.subscription_id !== 'string') {
        return [false, 'params.subscription_id must be a string.'];
    }
    if (params.address === undefined && params.topics === undefined) {
        return [false, 'params.address or params.topics must be provided.'];
    }
    if (params.address !== undefined && !Array.isArray(params.address)) {
        return [false, 'params.address must be an array of strings.'];
    }
    if (params.address !== undefined && params.address.some((addr) => typeof addr !== 'string')) {
        return [false, 'params.address must be an array of strings.'];
    }
    if (params.topics !== undefined && !Array.isArray(params.topics)) {
        return [false, 'params.topics must be an array.'];
    }
    if (params.topics !== undefined && params.topics.some((topic) => typeof topic !== 'string')) {
        return [false, 'params.topics must be an array of strings.'];
    }
    return [true, ''];
};
const validateUnsubscribeRequest = (request) => {
    if (typeof request !== 'object' || request === null) {
        return [false, 'Request must be an object.'];
    }
    const req = request;
    if (typeof req.params !== 'object' || req.params === null) {
        return [false, 'params must be an object.'];
    }
    const params = req.params;
    if (typeof params.subscription_id !== 'string') {
        return [false, 'params.subscription_id must be a string.'];
    }
    return [true, ''];
};
exports.evmLogSubscriptionHandler = {
    onMessage: async function (conn, message, socketId) {
        const [idValid, errorMsg] = validateBaseRequest(message);
        if (!idValid) {
            console.log(`Invalid request: ${errorMsg}, message: ${JSON.stringify(message)}`);
            conn.socket.send(JSON.stringify({ success: false, error: errorMsg }));
            return;
        }
        const request = message;
        switch (request.method) {
            case 'subscribe': {
                const [idValid, errorMsg] = validateSubscribeRequest(request);
                if (!idValid) {
                    console.log(`Invalid subscribe request: ${errorMsg}, message: ${JSON.stringify(message)}`);
                    conn.socket.send(JSON.stringify({ method: 'subscribe', success: false, error: errorMsg }));
                    return;
                }
                const subscribeRequest = request;
                (0, SocketManager_1.addLogSubscriptions)(subscribeRequest.params.subscription_id, socketId, {
                    address: Array.from(subscribeRequest.params.address || []).map((addr) => addr.toLowerCase()),
                    topics: Array.from(subscribeRequest.params.topics || []).map((topic) => topic.toLowerCase()),
                });
                conn.socket.send(JSON.stringify({
                    method: 'subscribe',
                    success: true,
                    subscription_id: subscribeRequest.params.subscription_id,
                }));
                (0, SocketManager_1.addLogSocketClient)(socketId, conn);
                return;
            }
            case 'unsubscribe': {
                const [idValid, errorMsg] = validateUnsubscribeRequest(request);
                if (!idValid) {
                    console.log(`Invalid unsubscribe request: ${errorMsg}, message: ${JSON.stringify(message)}`);
                    conn.socket.send(JSON.stringify({ method: 'unsubscribe', success: false, error: errorMsg }));
                    return;
                }
                const unsubscribeRequest = request;
                (0, SocketManager_1.removeLogSubscription)(unsubscribeRequest.params.subscription_id);
                conn.socket.send(JSON.stringify({
                    method: 'unsubscribe',
                    success: true,
                    subscription_id: unsubscribeRequest.params.subscription_id,
                }));
                return;
            }
            default:
                conn.socket.send(JSON.stringify({ success: false, error: 'invalid method' }));
                return;
        }
    },
};
//# sourceMappingURL=Handler.js.map