"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.usageMetricsHandler = exports.usageDisableHandler = exports.usageEnableHandler = exports.usageErrorMiddleware = exports.usageMiddleware = void 0;
const moment_1 = __importDefault(require("moment"));
const config_1 = require("../config");
const usageMetrics = {
    enabled: false,
    enabledAt: null,
    counter: {},
    errorsCounter: {},
};
const validateSecurityKey = (req, reply) => {
    const securityKey = req.headers['x-usage-key'];
    if (config_1.config.USAGE_ENDPOINTS_KEY === securityKey) {
        return true;
    }
    reply.code(403);
    reply.send('Unauthorized');
    return false;
};
const validUrl = (url) => {
    return url.startsWith('/api');
};
const ignoreQueryValues = (url) => {
    if (url.includes('?')) {
        const split_query = url.split('&');
        let final_url = '';
        for (const query of split_query) {
            if (query.includes('=')) {
                if (final_url != '')
                    final_url += '&' + query.substring(0, query.indexOf('='));
                else
                    final_url += query.substring(0, query.indexOf('='));
            }
        }
        return final_url;
    }
    return url;
};
/**
 * Usage middleware, intercepts requests and counts them by endpoint, only if usage metrics are enabled.
 */
const usageMiddleware = (req, _, done) => {
    if (!usageMetrics.enabled) {
        done();
        return;
    }
    let url = req.raw.url;
    if (!validUrl(url)) {
        done();
        return;
    }
    url = ignoreQueryValues(url);
    /* eslint-disable security/detect-object-injection */
    usageMetrics.counter[url] = (usageMetrics.counter[url] || 0) + 1;
    done();
};
exports.usageMiddleware = usageMiddleware;
/**
 * Usage error middleware, intercepts errors and counts them by endpoint and error message, only if usage metrics are enabled.
 */
const usageErrorMiddleware = (req, _, error, done) => {
    if (!usageMetrics.enabled) {
        done();
        return;
    }
    const url = req.raw.url;
    /* eslint-disable security/detect-object-injection */
    const endpointErrors = usageMetrics.errorsCounter[url] || {};
    endpointErrors[error.message] = (endpointErrors[error.message] || 0) + 1;
    usageMetrics.errorsCounter[url] = endpointErrors;
    done();
};
exports.usageErrorMiddleware = usageErrorMiddleware;
/**
 * Usage enable handler, enables usage metrics and resets the counters.
 */
const usageEnableHandler = async (req, reply) => {
    if (!validateSecurityKey(req, reply)) {
        return;
    }
    if (!usageMetrics.enabled) {
        usageMetrics.enabled = true;
        usageMetrics.enabledAt = new Date();
        usageMetrics.counter = {};
        usageMetrics.errorsCounter = {};
    }
    reply.send({
        status: 'enabled',
        enabledAt: usageMetrics.enabledAt,
    });
};
exports.usageEnableHandler = usageEnableHandler;
/**
 * Usage disable handler, disables usage metrics and resets the counters.
 */
const usageDisableHandler = async (req, reply) => {
    if (!validateSecurityKey(req, reply)) {
        return;
    }
    usageMetrics.enabled = false;
    usageMetrics.enabledAt = null;
    usageMetrics.counter = {};
    usageMetrics.errorsCounter = {};
    reply.send({
        status: 'disabled',
    });
};
exports.usageDisableHandler = usageDisableHandler;
/**
 * Usage metrics handler, returns the usage metrics.
 */
const usageMetricsHandler = async (req, reply) => {
    if (!validateSecurityKey(req, reply)) {
        return;
    }
    if (!usageMetrics.enabled) {
        reply.send({
            enabled: usageMetrics.enabled,
        });
    }
    reply.send({
        enabled: usageMetrics.enabled,
        enabledAt: usageMetrics.enabledAt,
        enableForInMinutes: (0, moment_1.default)(new Date()).diff((0, moment_1.default)(usageMetrics.enabledAt), 'minutes'),
        usage: usageMetrics.counter,
        errors: usageMetrics.errorsCounter,
    });
};
exports.usageMetricsHandler = usageMetricsHandler;
//# sourceMappingURL=usage.js.map